@@ -32,8 +32,7 @@ class XQuerySet
 		char * header;
 		char * text;
 		XQuerySet ** qs;
-		bool global_and; // global
-		bool global_neg; // global
+		Xapian::Query::op global_op;
 		bool item_neg; // for the term
 		long qsize;
 		long limit;
@@ -42,22 +41,20 @@ class XQuerySet
 	{
 		qsize=0; qs=NULL;
 		limit=1;
-		global_and=true;
 		header=NULL;
 		text=NULL;
-		global_neg=false;
+		global_op = Xapian::Query::op::OP_OR;		
 		accentsConverter=NULL;
 	}
 
-	XQuerySet(bool is_and, bool is_neg, long l)
+	XQuerySet(Xapian::Query::op op, long l)
 	{
 		qsize=0; qs=NULL;
 		limit=1;
 		if(l>1) { limit=l; }
 		header=NULL;
 		text=NULL;
-		global_and=is_and;
-		global_neg=is_neg;
+		global_op=op;
 		accentsConverter=NULL;
 	}
 
@@ -93,69 +90,69 @@ class XQuerySet
 
 	void add(icu::UnicodeString *h, icu::UnicodeString *t, bool is_neg)
 	{
-		long i,j;
-		XQuerySet * q2;
-		icu::UnicodeString *r;
-
-		t->findAndReplace("'"," ");
-		t->findAndReplace("\""," ");
-		t->findAndReplace(":"," ");
-		t->findAndReplace(";"," ");
-		t->findAndReplace("\""," ");
-		t->findAndReplace("<"," ");
-		t->findAndReplace(">"," ");
-		t->findAndReplace("\n"," ");
-		t->findAndReplace("\r"," ");
-		t->findAndReplace("@"," ");
-		t->findAndReplace("-","_");
-
 		h->trim();
-		t->trim();
-		h->toLower();
+                h->toLower();
+                if(h->length()<1) return;
+
+		long i,j,k;
+		XQuerySet * q2;
+		icu::UnicodeString *r1,*r2;
+		std::string st1,st2;
+
+		t->findAndReplace("'",CHAR_SPACE);
+		t->findAndReplace("\"",CHAR_SPACE);
+		t->findAndReplace("\n",CHAR_SPACE);
+		t->findAndReplace("\r",CHAR_SPACE);
+		t->findAndReplace("\t",CHAR_SPACE);
 		t->toLower();
 
-		if(h->length()<1) return;
+		k=CHARS_PB;
+		while(k>0)
+		{
+			t->findAndReplace(chars_pb[k-1],CHAR_KEY);
+			k--;
+		}
+		while((t->indexOf(CHAR_KEY)==0) || (t->indexOf(CHAR_SPACE)==0))
+                {
+                        t->remove(0,1);
+                }
+                i = std::max(t->lastIndexOf(CHAR_KEY),t->lastIndexOf(CHAR_SPACE));
+                while((i>0) && (i==t->length()-1))
+                {
+                        t->remove(i,1);
+                        i = std::max(t->lastIndexOf(CHAR_KEY),t->lastIndexOf(CHAR_SPACE));
+                }
 		if(t->length()<limit) return;
-
-		i = t->lastIndexOf(" ");
-		if(i>0)
-		{
-			q2 = new XQuerySet(true,false,limit);
-			while(i>0)
+	
+		i = t->lastIndexOf(CHAR_SPACE);
+                if(i>0)
+                {
+			if(is_neg)
 			{
-				j = t->length();
-				r = new icu::UnicodeString(*t,i+1,j-i-1);
-				q2->add(h,r,false);
-				delete(r);
-				t->truncate(i);
-				t->trim();
-				i = t->lastIndexOf(" ");
+                        	q2 = new XQuerySet(Xapian::Query::OP_AND_NOT,limit);
 			}
-			q2->add(h,t,false);
-			if(q2->count()>0) add(q2); else delete(q2);
-			return;
-		}
+			else
+			{
+				q2 = new XQuerySet(Xapian::Query::OP_AND,limit);
+			}
+                        while(i>0)
+                        {
+                                j = t->length();
+                                r1 = new icu::UnicodeString(*t,i+1,j-i-1);
+                                q2->add(h,r1,false);
+                                delete(r1);
+                                t->truncate(i);
+                                t->trim();
+                                i = t->lastIndexOf(CHAR_SPACE);
+                        }
+                        q2->add(h,t,false);
+                        if(q2->count()>0) add(q2); else delete(q2);
+                        return;
+                }
 
-		i = t->indexOf(".");
-		if(i>=0)
-		{
-			r = new icu::UnicodeString(*t);
-			r->findAndReplace(".","_");
-			q2 = new XQuerySet(false,false,limit);
-			q2->add(h,r,false);
-			delete(r);
-
-			t->findAndReplace("."," ");
-			t->trim();
-			q2->add(h,t,false);
-
-			if(q2->count()>0) add(q2); else delete(q2);
-			return;
-		}
-
-		std::string tmp1;
-		h->toUTF8String(tmp1);
-		char * h2 = i_strdup(tmp1.c_str());
+		st1.clear();
+		h->toUTF8String(st1);
+		char * h2 = i_strdup(st1.c_str());
 
 		if(accentsConverter == NULL)
 		{
@@ -169,16 +166,23 @@ class XQuerySet
 		}
 		if(accentsConverter != NULL) accentsConverter->transliterate(*t);
 
-		std::string tmp2;
-		t->toUTF8String(tmp2);
-		char * t2 = i_strdup(tmp2.c_str());
+		st2.clear();
+		t->toUTF8String(st2);
+		char * t2 = i_strdup(st2.c_str());
 
 		if(strcmp(h2,XAPIAN_WILDCARD)==0)
 		{
-			q2 = new XQuerySet(false,is_neg,limit);
+			if(is_neg)
+			{
+				q2 = new XQuerySet(Xapian::Query::OP_AND_NOT,limit);
+			}
+			else
+			{
+                        	q2 = new XQuerySet(Xapian::Query::OP_OR,limit);
+			}
 			for(i=1;i<HDRS_NB;i++)
 			{
-				if(i!=XAPIAN_EXPUNGE_HEADER) q2->add(hdrs_emails[i],t2,is_neg);
+				if(i!=XAPIAN_EXPUNGE_HEADER) q2->add(hdrs_emails[i],t2,false);
 			}
 			add(q2);
 			i_free(h2);
@@ -198,13 +202,6 @@ class XQuerySet
 			return;
 		}
 
-		if(has(h2,t2,true))
-		{
-			i_free(h2);
-			i_free(t2);
-			return;
-		}
-
 		if(text==NULL)
 		{
 			text=t2;
@@ -213,13 +210,14 @@ class XQuerySet
 			return;
 		}
 
-		q2 = new XQuerySet(global_and,is_neg,limit);
-		q2->add(h,t,false);
+		q2 = new XQuerySet(Xapian::Query::OP_AND,limit);
+		q2->add(h,t,is_neg);
 		add(q2);
 	}
 
 	void add(XQuerySet *q2)
 	{
+		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: XQuerySet->addQ()");
 		if(qsize<1)
 		{
 			qs=(XQuerySet **)i_malloc(sizeof(XQuerySet*));
@@ -232,19 +230,6 @@ class XQuerySet
 		qsize++;
 	}
 
-	bool has(const char *h, const char *t, bool loop)
-	{
-		if((text!=NULL) && (strcmp(h,header)==0) && (strcmp(t,text)==0)) return true;
-		if(loop)
-		{
-			for(long i=0; i<qsize; i++)
-			{
-				if(qs[i]->has(h,t,false)) return true;
-			}
-		}
-		return false;
-	}
-
 	int count()
 	{
 		int c=0;
@@ -255,23 +240,29 @@ class XQuerySet
 
 	std::string get_string()
 	{
-		std::string s;
+		std::string s("");
 
 		if(count()<1) return s;
 
 		if(text!=NULL)
 		{
-			if(item_neg) s.append("NOT( ");
+			if(item_neg) s.append("NOT ( ");
 			s.append(header);
-			//s.append(":");
-			s.append(":\"");
+			s.append(":");
+			s.append("\"");
 			s.append(text);
 			s.append("\"");
-			if(item_neg) s.append(" )");
+			if(item_neg) s.append(")");
 		}
 
-		const char * op=" OR ";
-		if(global_and) op=" AND ";
+		const char * op;
+		switch(global_op)
+		{
+			case Xapian::Query::OP_OR : op=" OR "; break;
+			case Xapian::Query::OP_AND : op=" AND "; break;
+			case Xapian::Query::OP_AND_NOT : op=" AND NOT "; break;
+			default : op=" ERROR ";
+		}	
 
 		for (int i=0;i<qsize;i++)
 		{
@@ -280,13 +271,7 @@ class XQuerySet
 
 			if(s.length()>0) s.append(op);
 
-			if(qs[i]->global_neg)
-			{
-				s.append("NOT(");
-				s.append(qs[i]->get_string());
-				s.append(")");
-			}
-			else if(c>1)
+			if(c>1)
 			{
 				s.append("(");
 				s.append(qs[i]->get_string());
@@ -294,32 +279,58 @@ class XQuerySet
 			}
 			else s.append(qs[i]->get_string());
 		}
-		if(global_neg)
-		{
-			s="NOT("+s+")";
-		}
 		return s;
 	}
 
 	Xapian::Query * get_query(Xapian::Database * db)
 	{
-		if(count()<1)
-		{
-			return new Xapian::Query(Xapian::Query::MatchNothing);
+		Xapian::Query * q = NULL;
+		Xapian::Query *q2, *q3;
+
+		if(text!=NULL)
+                {
+			std::string s(header);
+                        s.append(":");
+                        s.append("\"");
+                        s.append(text);
+                        s.append("\"");
+
+			Xapian::QueryParser * qp = new Xapian::QueryParser();
+			for(int i=0; i< HDRS_NB; i++) qp->add_prefix(hdrs_emails[i], hdrs_xapian[i]);
+			qp->set_database(*db);
+			q = new Xapian::Query(qp->parse_query(s.c_str(),Xapian::QueryParser::FLAG_DEFAULT));
+			delete (qp);
+			if(item_neg)
+			{
+				q2 = new Xapian::Query(Xapian::Query::MatchAll);
+				q3 = new Xapian::Query(Xapian::Query::OP_AND_NOT,*q2,*q);
+				delete(q2);
+				delete(q);
+				q=q3;
+			}
 		}
+		if(qsize<1)
+		{
+			if(q==NULL) q = new Xapian::Query(Xapian::Query::MatchNothing);
+			return q;
+                }
 
-		Xapian::QueryParser * qp = new Xapian::QueryParser();
-
-		for(int i=0; i< HDRS_NB; i++) qp->add_prefix(hdrs_emails[i], hdrs_xapian[i]);
-
-		char *s = i_strdup(get_string().c_str());
-
-		qp->set_database(*db);
-
-		Xapian::Query * q = new Xapian::Query(qp->parse_query(s,Xapian::QueryParser::FLAG_DEFAULT));// | Xapian::QueryParser::FLAG_PARTIAL));
-
-		i_free(s);
-		delete(qp);
+		if(q==NULL)
+		{	
+			q=qs[0]->get_query(db);
+		}
+		else
+		{
+			q2 = new Xapian::Query(global_op,*q,*(qs[0]->get_query(db)));
+			delete(q);
+			q=q2;
+		}
+		for (int i=1;i<qsize;i++)
+		{
+			q2 = new Xapian::Query(global_op,*q,*(qs[i]->get_query(db)));
+			delete(q);
+			q=q2;
+		}
 		return q;
 	}
 };
@@ -375,83 +386,107 @@ class XNGram
 
 	void add(icu::UnicodeString *d)
 	{
-		icu::UnicodeString * r;
+		if(fts_xapian_settings.verbose>2) i_info("FTS Xapian: XNGram->add()");
 
 		d->toLower();
-		d->findAndReplace("'"," ");
-		d->findAndReplace("\""," ");
-		d->findAndReplace(":"," ");
-		d->findAndReplace(";"," ");
-		d->findAndReplace("\""," ");
-		d->findAndReplace("<"," ");
-		d->findAndReplace(">"," ");
-		d->findAndReplace("\n"," ");
-		d->findAndReplace("\r"," ");
-		d->findAndReplace("@"," ");
-		d->findAndReplace("-","_");
-		
-		long i = d->indexOf(".");
-		if(i>=0)
-		{
-			r = new icu::UnicodeString(*d);
-			r->findAndReplace(".","_");
-			add(r);
-			delete(r);
-			d->findAndReplace("."," ");
-		}
-
+		d->findAndReplace("'",CHAR_SPACE);
+		d->findAndReplace("\"",CHAR_SPACE);
+		d->findAndReplace("\n",CHAR_SPACE);
+		d->findAndReplace("\r",CHAR_SPACE);
+		d->findAndReplace("\t",CHAR_SPACE);
 		d->trim();
-		i = d->indexOf(" ");
+	
+		long i,k;
+		icu::UnicodeString *r1,*r2;
+                i = d->lastIndexOf(CHAR_SPACE);
 
+                if(i>0)
+                {
+                        r1 = new icu::UnicodeString(*d,0,i);
+                        r2 = new icu::UnicodeString(*d,i+1,d->length()-i-1);
+                        add(r1);
+                        add(r2);
+                        delete(r1);
+                        delete(r2);
+                        return;
+                }
+
+		k=CHARS_PB;
+                while(k>0)
+                {
+			d->findAndReplace(chars_pb[k-1],CHAR_KEY);
+			k--;
+		}
+                
+		if(accentsConverter == NULL)
+                {
+                        UErrorCode status = U_ZERO_ERROR;
+                        accentsConverter = icu::Transliterator::createInstance("NFD; [:M:] Remove; NFC", UTRANS_FORWARD, status);
+                        if(U_FAILURE(status))
+                        {
+                                i_error("FTS Xapian: Can not allocate ICU translator (1)");
+                                accentsConverter = NULL;
+                        }
+                }
+                if(accentsConverter != NULL) accentsConverter->transliterate(*d);
+       
+		bool b=false;
+		while(d->indexOf(CHAR_KEY)==0)
+		{
+			d->remove(0,1);
+			b=true;
+		}
+		i = d->lastIndexOf(CHAR_KEY);
+		while((i>0) && (i==d->length()-1))
+		{
+			d->remove(i,1);
+			i = d->lastIndexOf(CHAR_KEY);
+			b=true;
+		}
+	
+                k = d->length();
+                if(k<fts_xapian_settings.partial) return;
+
+                if(onlyone)
+                {
+                        add_stem(d);
+                        return;
+                }
+	
 		if(i>0)
 		{
-			r = new icu::UnicodeString(*d,i+1);
-			add(r);
-			delete(r);
-			d->truncate(i);
-			d->trim();
-		}
-
-		long l = d->length();
-		if(l<fts_xapian_settings.partial) return;
-
-		if(accentsConverter == NULL)
-		{
-			UErrorCode status = U_ZERO_ERROR;
-			accentsConverter = icu::Transliterator::createInstance("NFD; [:M:] Remove; NFC", UTRANS_FORWARD, status);
-			if(U_FAILURE(status))
-			{
-				i_error("FTS Xapian: Can not allocate ICU translator (1)");
-				accentsConverter = NULL;
-			}
-		}
-		if(accentsConverter != NULL) accentsConverter->transliterate(*d);
-
-		if(onlyone)
-		{
+			r1 = new icu::UnicodeString(*d,0,i);
+                        r2 = new icu::UnicodeString(*d,i+1,d->length()-i-1);
+			add(r1);
+			add(r2);
+			delete(r1);
+			delete(r2);
 			add_stem(d);
 			return;
 		}
 
-		for(i=0;i<=l-fts_xapian_settings.partial;i++)
+		for(i=0;i<=k-fts_xapian_settings.partial;i++)
 		{
-			for(long j=fts_xapian_settings.partial;(j+i<=l)&&(j<=fts_xapian_settings.full);j++)
+			for(long j=fts_xapian_settings.partial;(j+i<=k)&&(j<=fts_xapian_settings.full);j++)
 			{
-				r = new icu::UnicodeString(*d,i,j);
-				add_stem(r);
-				delete(r);
+				r1 = new icu::UnicodeString(*d,i,j);
+				add_stem(r1);
+				delete(r1);
 			}
 		}
-		if(l>fts_xapian_settings.full) add_stem(d);
+		if(b || (k>fts_xapian_settings.full)) add_stem(d);
 	}
 
 	void add_stem(icu::UnicodeString *d)
 	{
+		long l,i,p;
+		std::string s;
+		char * s2;
+
 		d->trim();
-		long l=d->length();
+		l=d->length();
 		if(l<fts_xapian_settings.partial) return;
 
-		std::string s;
 		d->toUTF8String(s);
 		l  = s.length();
 		if(l>hardlimit)
@@ -460,8 +495,10 @@ class XNGram
 			return;
 		}
 
-		char * s2 = i_strdup(s.c_str());
-		long p =0;
+		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: XNGram->add_stem(%s)",s.substr(0,100).c_str());
+
+		s2 = i_strdup(s.c_str());
+		p=0;
 
 		if(size<1)
 		{
@@ -482,7 +519,7 @@ class XNGram
 				return;
 			}
 			data=(char **)i_realloc(data,size*sizeof(char*),(size+1)*sizeof(char*));
-			long i=size;
+			i=size;
 			while(i>p)
 			{
 				data[i]=data[i-1];
@@ -505,42 +542,49 @@ static long fts_backend_xapian_current_time()
 
 static long fts_backend_xapian_get_free_memory() // KB
 {
-	struct rlimit rl;
-	getrlimit(RLIMIT_AS,&rl);
-
-	long limit = rl.rlim_cur / 1024.0;
-	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: RLIM AS =%ld",limit);
-
-	getrlimit(RLIMIT_DATA,&rl);
-	long l2 = rl.rlim_cur / 1024.0;
-        if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: RLIM DATA =%ld",l2);
-
-	if((l2>0) && ((limit>l2) || (limit<1))) limit=l2;
-
-#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__APPLE__)	
-	uint32_t m;
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__)	
+	uint32_t m,n;
 	size_t len = sizeof(m);
 	sysctlbyname("vm.stats.vm.v_free_count", &m, &len, NULL, 0);
 	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: (BSD) Free pages %ld",long(m));
-	m = m * fts_xapian_settings.pagesize / 1024.0;
+	sysctlbyname("vm.stats.vm.v_cache_count", &n, &len, NULL, 0);
+	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: (BSD) Cached pages %ld",long(n));
+	m = (m+n) * fts_xapian_settings.pagesize / 1024.0;
 	if((limit>0) && (m>limit)) m = limit;
-	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: (BSD) Free memory %ld MB",long(m/1024.0));
+	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: (BSD) Available memory %ld MB",long(m/1024.0));
 	return long(m);
 #else
-	long m = long(sysconf(_SC_AVPHYS_PAGES) * fts_xapian_settings.pagesize / 1024.0);
-	if((limit>0) && (m>limit)) m = limit;
-	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: Free memory %ld MB",long(m/1024.0));
-	return m;
+        long m=0;
+        char buffer[500];
+        const char *p;
+        FILE *f=fopen("/proc/meminfo","r");
+        while(!feof(f))
+        {       
+                if ( fgets (buffer , 100 , f) == NULL ) break;
+                p = strstr(buffer,"MemFree");
+                if(p!=NULL)
+                {
+                        m+=atol(p+8);
+                }
+                p = strstr(buffer,"Cached");
+                if(p==buffer)
+                {
+                        m+=atol(p+7);
+                }
+        }       
+        if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: Free memory %ld MB",long(m/1024.0));
+        fclose (f);     
+        return m;
 #endif
 }
 
-static bool fts_backend_xapian_test_memory()
+static long fts_backend_xapian_test_memory()
 {
 	long fri = fts_backend_xapian_get_free_memory(); // KB
 
-	if(fts_xapian_settings.verbose>1) i_warning("FTS Xapian: Free memory %ld MB vs %ld MB minimum",long(fri/1024.0),fts_xapian_settings.lowmemory);
-
-	return(fri > ( fts_xapian_settings.lowmemory * 1024 ) );
+	if(fri < ( fts_xapian_settings.lowmemory * 1024L ) ) return fri;
+	
+	return -1L;
 }
 
 static bool fts_backend_xapian_open_readonly(struct xapian_fts_backend *backend, Xapian::Database ** dbr)
@@ -588,17 +632,17 @@ static bool fts_backend_xapian_check_access(struct xapian_fts_backend *backend)
 		i_error("FTS Xapian: Can't open Xapian DB (RW) (%s) %s : %s - %s",backend->boxname,backend->db,e.get_type(),e.get_error_string());
 		return false;
 	}
+	backend->nbdocs=backend->dbw->get_doccount();
 	if(fts_xapian_settings.verbose>0) 
 	{
-		long n = backend->dbw->get_doccount();
-		i_info("FTS Xapian: Opening DB (RW) %s (%ld docs stored): Done",backend->db,n);
+		i_info("FTS Xapian: Opening DB (RW) %s (%ld docs stored): Done",backend->db,backend->nbdocs);
 	}
 	return true;
 }
 
 static void fts_backend_xapian_oldbox(struct xapian_fts_backend *backend)
 {
-	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_oldbox");
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_oldbox");
 
 	if(backend->old_guid != NULL)
 	{
@@ -612,13 +656,70 @@ static void fts_backend_xapian_oldbox(struct xapian_fts_backend *backend)
 		}
 		/* End Performance calculator*/
 
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Done indexing '%s' (%s) (%ld msgs in %ld ms, rate: %.1f)",backend->old_boxname, backend->old_guid,backend->perf_nb,dt,r);
+		i_info("FTS Xapian: Done indexing '%s' (%s) (%ld msgs in %ld ms, rate: %.1f)",backend->old_boxname, backend->old_guid,backend->perf_nb,dt,r);
 
 		i_free(backend->old_guid); backend->old_guid = NULL;
 		i_free(backend->old_boxname); backend->old_boxname = NULL;
 	}
+
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_oldbox - done");
 }
 
+static std::string fts_backend_xapian_get_selfpath() 
+{
+    char buff[PATH_MAX];
+    buff[0]=0;
+    ssize_t len = ::readlink("/proc/self/exe", buff, sizeof(buff)-1);
+    if (len != -1) 
+    {
+      buff[len] = '\0';
+    }
+    return std::string(buff);
+}
+
+static void fts_backend_xapian_commitclose(Xapian::WritableDatabase * db, long nbdocs, std::string * dbpath, std::string * boxname, bool threaded)
+{
+	const char * c="Not threaded";
+	if(threaded) c="Threaded";
+
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian (%s): Commit & Closing (%s) starting (%s) : %s",c,boxname->c_str(),dbpath->c_str(),fts_backend_xapian_get_selfpath().c_str());
+	bool err=false;
+	long n = 0;
+	long t = fts_backend_xapian_current_time();
+	if(fts_xapian_settings.verbose>0) n = db->get_doccount();
+	try
+	{
+		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian (%s): Commit & Closing (%s) : Writing %ld (old) vs %ld (new)",c,boxname->c_str(),nbdocs,n);	
+		db->commit();
+		db->close();
+	}
+	catch(Xapian::Error e)
+        {
+        	i_error("FTS Xapian: %s - %s",e.get_type(),e.get_error_string());
+                err=true;
+        }
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian (%s): Commit & Closing (%s) : Releasing Xapian db",c,boxname->c_str());
+	delete(db);
+	t = fts_backend_xapian_current_time() -t;
+        if(err)
+        {
+        	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Re-creating index database (%s) due to error",dbpath->c_str());
+                try
+                {
+                	db = new Xapian::WritableDatabase(dbpath->c_str(),Xapian::DB_CREATE_OR_OVERWRITE | Xapian::DB_RETRY_LOCK | Xapian::DB_BACKEND_GLASS | Xapian::DB_NO_SYNC);
+                        db->close();
+                        delete(db);
+                }
+                catch(Xapian::Error e)
+                {
+                        i_error("FTS Xapian: Can't re-create Xapian DB %s : %s - %s",dbpath->c_str(),e.get_type(),e.get_error_string());
+                }
+        }
+	else if(fts_xapian_settings.verbose>0) i_info("FTS Xapian (%s) : Commit & Close (%s) (%s) - Done in %ld ms",c,dbpath->c_str(),boxname->c_str(),t);
+	delete(dbpath);
+	delete(boxname);
+}
+	
 static void fts_backend_xapian_release(struct xapian_fts_backend *backend, const char * reason, long commit_time)
 {
 	bool err=false;
@@ -627,69 +728,47 @@ static void fts_backend_xapian_release(struct xapian_fts_backend *backend, const
 
 	if(commit_time<1) commit_time = fts_backend_xapian_current_time();
 
-	long n = 0;
 	if(backend->dbw !=NULL)
 	{
-		if(fts_xapian_settings.verbose>0) n = backend->dbw->get_doccount();
-		try
+		std::string *a = new std::string(backend->db);
+		std::string *b = new std::string(backend->boxname);
+
+		if(strstr(fts_backend_xapian_get_selfpath().c_str(),"doveadm")==NULL)
 		{
-			backend->dbw->commit();
-			backend->dbw->close();
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian - Lauching Thread for closing");
+			try
+			{
+				new std::thread(fts_backend_xapian_commitclose,backend->dbw,backend->nbdocs,a,b,true);
+			}
+			catch (std::runtime_error &ex)
+			{
+				if(fts_xapian_settings.verbose>0) i_info("FTS Xapian - Thread Closing ERROR(%s)",ex.what());
+			}
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian - Thread for closing launched");
 		}
-		catch(Xapian::Error e)
+		else
 		{
-			i_error("FTS Xapian: %s : %s - %s",reason,e.get_type(),e.get_error_string());
-			err=true;
+			fts_backend_xapian_commitclose(backend->dbw,backend->nbdocs,a,b,false);
 		}
-		delete(backend->dbw);
 		backend->dbw = NULL;
 		backend->commit_updates = 0;
 		backend->commit_time = commit_time;
 	}
-
-	if(err)
-	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Re-creating index database due to error");
-		try
-		{
-			Xapian::WritableDatabase * db = new Xapian::WritableDatabase(backend->db,Xapian::DB_CREATE_OR_OVERWRITE | Xapian::DB_RETRY_LOCK | Xapian::DB_BACKEND_GLASS | Xapian::DB_NO_SYNC);
-			db->close();
-			delete(db);
-		}
-		catch(Xapian::Error e)
-		{
-			i_error("FTS Xapian: Can't re-create Xapian DB (%s) %s : %s - %s",backend->boxname,backend->db,e.get_type(),e.get_error_string());
-		}
-	}
-
-	if(fts_xapian_settings.verbose>0)
-	{
-		if(n>0)
-		{
-			i_info("FTS Xapian: Committed '%s' in %ld ms (%ld docs in index)",reason,fts_backend_xapian_current_time() - commit_time,n);
-		}
-		else
-		{
-			i_info("FTS Xapian: Committed '%s' in %ld ms",reason,fts_backend_xapian_current_time() - commit_time);
-		}
-	}
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_release (%s) - done",reason);
 }
 
 XResultSet * fts_backend_xapian_query(Xapian::Database * dbx, XQuerySet * query, long limit=0)
 {
-	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_query");
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_query (%s)",query->get_string().c_str());
 
 	XResultSet * set= new XResultSet();
+	Xapian::Query * q = query->get_query(dbx);
 
 	try
 	{
 		Xapian::Enquire enquire(*dbx);
-
-		Xapian::Query * q = query->get_query(dbx);
-
 		enquire.set_query(*q);
 		enquire.set_docid_order(Xapian::Enquire::DESCENDING);
-		delete(q);
 
 		long offset=0;
 		long pagesize=100; if(limit>0) { pagesize=std::min(pagesize,limit); }
@@ -711,6 +790,7 @@ XResultSet * fts_backend_xapian_query(Xapian::Database * dbx, XQuerySet * query,
 	{
 		i_error("FTS Xapian: xapian_query %s - %s",e.get_type(),e.get_error_string());
 	}
+	delete(q);
 	return set;
 }
 
@@ -841,6 +921,8 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 {
 	const char * hdr;
 
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_build_qs");
+
 	while(a != NULL)
 	{
 		switch (a->type)
@@ -870,7 +952,15 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 		}
 		if((a->value.str == NULL) || (strlen(a->value.str)<1))
 		{
-			XQuerySet * q2 = new XQuerySet(false,a->match_not,qs->limit);
+			XQuerySet * q2;
+			if(a->match_not)
+			{
+				q2 = new XQuerySet(Xapian::Query::OP_AND_NOT,qs->limit);
+			}
+			else
+			{
+				q2 = new XQuerySet(Xapian::Query::OP_OR,qs->limit);
+			}
 			fts_backend_xapian_build_qs(q2,a->value.subargs);
 			if(q2->count()>0)
 			{
@@ -1007,7 +1097,7 @@ bool fts_backend_xapian_index_text(struct xapian_fts_backend *backend,uint uid,
 {
 	bool ok = true;
 
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_index_text");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_index_text");
 
 	Xapian::WritableDatabase * dbx = backend->dbw;
 
@@ -1089,7 +1179,7 @@ bool fts_backend_xapian_index_text(struct xapian_fts_backend *backend,uint uid,
 		n--;
 	}
 
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: NGRAM(%s,%s) -> %ld items, max length=%ld, (total %ld KB)",field,h,ngram->size,ngram->maxlength,ngram->memory/1024);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: NGRAM(%s,%s) -> %ld items, max length=%ld, (total %ld KB)",field,h,ngram->size,ngram->maxlength,ngram->memory/1024);
 
 	char *t = (char*)i_malloc(sizeof(char)*(ngram->maxlength+6));
 	for(n=0;n<ngram->size;n++)
