@@ -1,6 +1,7 @@
 /* Copyright (c) 2019 Joan Moreau <jom@grosjo.net>, see the included COPYING file */
 
 #include <xapian.h>
+#include <thread>
 #include <cstdio>
 #include <vector>
 extern "C" {
@@ -17,6 +18,10 @@ static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "c
 static const char * hdrs_xapian[HDRS_NB] = { "Q",   "S",       "A",    "XTO", "XCC", "XBCC", "XMID",      "XLIST",  "XBDY", "XBDY" };
 static const char * createTable = "CREATE TABLE IF NOT EXISTS docs(ID INT PRIMARY KEY NOT NULL);";
 static const char * selectUIDs = "select ID from docs;";
+#define CHARS_PB 18
+#define CHAR_KEY "_"
+#define CHAR_SPACE " "
+static const char * chars_pb[] = { "<", ">", ".", "-", "@", ":", ";", "!", "%", "*", "|", "`", "#", "~", "^", ",", "/", "\\" };
 
 struct xapian_fts_backend
 {
@@ -40,6 +45,8 @@ struct xapian_fts_backend
 	long perf_nb;
 	long perf_uid;
 	long perf_dt;
+
+	long nbdocs = 0;
 };
 
 struct xapian_fts_backend_update_context
@@ -84,7 +91,7 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 
 	if(fts_backend_xapian_set_path(backend)<0) return -1;
 
-        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld verbose=%d lowmemory=%ld MB vs freemem=%ld MB",fts_xapian_settings.partial,fts_xapian_settings.full,fts_xapian_settings.verbose,fts_xapian_settings.lowmemory, long(fts_backend_xapian_get_free_memory()/1024.0));
+        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld verbose=%d lowmemory=%ld MB vs freemem=%ld MB", fts_xapian_settings.partial,fts_xapian_settings.full,fts_xapian_settings.verbose,fts_xapian_settings.lowmemory, long(fts_backend_xapian_get_free_memory()/1024.0));
 
 	return 0;
 }
@@ -417,9 +424,11 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 		return -1;
 	}
 
-	if(!fts_backend_xapian_test_memory())
+	long fri = fts_backend_xapian_test_memory();
+
+	if(fri>=0)
 	{
-		if(fts_xapian_settings.verbose>0) i_warning("FTS Xapian: Warning Low memory (%ld MB)",long(fts_backend_xapian_get_free_memory()/1024.0));
+		i_warning("FTS Xapian: Warning Free memory %ld MB < %ld MB minimum",long(fri/1024.0),fts_xapian_settings.lowmemory);
 		fts_backend_xapian_release(backend,"Low memory indexing", 0);
 		if(!fts_backend_xapian_check_access(backend))
 		{
@@ -515,6 +524,9 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 	int ret=0;
 	std::vector<uint32_t> uids(0);
 	char *zErrMsg = 0;
+	XResultSet * result = NULL;
+	Xapian::docid docid =0;
+
 	while ((dp = readdir(dirp)) != NULL)
 	{
 		if((dp->d_type == DT_DIR) && (strncmp(dp->d_name,"db_",3)==0))
@@ -543,7 +555,7 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 				}
 				i_free(s);
 				s = i_strdup_printf("%s/%s",backend->path,dp->d_name);
-				if(fts_xapian_settings.verbose>0) i_info("Optimize (5a) Opening Xapian DB (%s)",s);
+				if(fts_xapian_settings.verbose>0) i_info("Optimize (4) Opening Xapian DB (%s)",s);
 				try
 				{
 					db = new Xapian::WritableDatabase(s,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK | Xapian::DB_BACKEND_GLASS | Xapian::DB_NO_SYNC);
@@ -551,37 +563,35 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 					{
 						uid=uids[n];
 						if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (5) Removing DOC UID=%d",uid);
-						char * d = i_strdup_printf("delete from docs where ID=%d",uid);
-						try
+						XQuerySet * xq = new XQuerySet();
+						char *u = i_strdup_printf("%d",uid);
+						xq->add("uid",u);
+						i_free(u);
+						result=fts_backend_xapian_query(db,xq,1);
+						docid=0;
+						if((result!=NULL) && (result->size>0))
 						{
-							XQuerySet * xq = new XQuerySet();
-        						char *u = i_strdup_printf("%d",uid);
-        						xq->add("uid",u);
-        						XResultSet * result=fts_backend_xapian_query(db,xq,1);
-							if(result->size>0)
+							try
 							{
-								Xapian::docid docid = result->data[0];
-								if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (5) Removing DOC UID=%d (%s) DOCID=%d",uid,d,docid);
+								docid = result->data[0];
+								if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (5) Removing DOC UID=%d DOCID=%d",uid,docid);
 								db->delete_document(docid);
-								if (sqlite3_exec(expdb,d,NULL,0,&zErrMsg) != SQLITE_OK )
-								{
-									i_error("FTS Xapian : Optimize Sqlite error %s",zErrMsg);
-									sqlite3_free(zErrMsg);
-								}
 							}
-							else
+							catch(Xapian::Error e)
 							{
-								i_error("FTS Xapian: Optimize UID=%d inexistant",uid);
+								i_error("FTS Xapian: Optimize (6) %s",e.get_msg().c_str());
 							}
-							delete(result);
-							i_free(u);
-							delete(xq);
 						}
-						catch(Xapian::Error e)
+						else if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize UID=%d (DOCID=%d) inexistent",uid,docid);
+						if(result!=NULL) { delete(result); result=NULL; }
+						delete(xq);
+						u = i_strdup_printf("delete from docs where ID=%d",uid);
+						if (sqlite3_exec(expdb,u,NULL,0,&zErrMsg) != SQLITE_OK )
 						{
-							i_error("FTS Xapian: Optimize (5a) %s",e.get_msg().c_str());
+							i_error("FTS Xapian : Optimize Sqlite error %s",zErrMsg);
+							sqlite3_free(zErrMsg);
 						}
-						i_free(d);	
+						i_free(u);
 					}
 					db->commit();
 					db->close();
@@ -671,19 +681,19 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 		return 0;
 	}
 
-	bool is_and=false;
+	XQuerySet * qs;
 
 	if((flags & FTS_LOOKUP_FLAG_AND_ARGS) != 0)
 	{
 		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: FLAG=AND");
-		is_and=true;
+		qs = new XQuerySet(Xapian::Query::OP_AND,fts_xapian_settings.partial);
 	}
 	else
 	{
 		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: FLAG=OR");
+		qs = new XQuerySet(Xapian::Query::OP_OR,fts_xapian_settings.partial);
 	}
 
-	XQuerySet * qs = new XQuerySet(is_and,false,fts_xapian_settings.partial);
 	fts_backend_xapian_build_qs(qs,args);
 
 	XResultSet * r=fts_backend_xapian_query(dbr,qs);
