@@ -7,8 +7,6 @@ const char *fts_xapian_plugin_version = DOVECOT_ABI_VERSION;
 
 struct fts_xapian_user_module fts_xapian_user_module = MODULE_CONTEXT_INIT(&mail_user_module_register);
 
-
-
 static void fts_xapian_mail_user_deinit(struct mail_user *user)
 {
 #if ((DOVECOT_VERSION_MINOR > 2) || (DOVECOT_VERSION_MAJOR > 2))
@@ -35,11 +33,9 @@ static void fts_xapian_mail_user_created(struct mail_user *user)
         fuser->set.partial 	= XAPIAN_DEFAULT_PARTIAL;
         fuser->set.full 	= XAPIAN_DEFAULT_FULL;
 
-#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__)
 	size_t len = sizeof(fuser->set.pagesize);
 	sysctlbyname("hw.pagesize", &(fuser->set.pagesize), &len, NULL, 0);
-#else
-	fuser->set.pagesize = sysconf(_SC_PAGE_SIZE);
 #endif
 
 	const char * env = mail_user_plugin_getenv(user, "fts_xapian");
